import os
import hnswlib
import numpy as np
import unittest


class RandomSelfTestCase(unittest.TestCase):
    def testRandomSelf(self):
        dim = 32
        num_elements = 100000
        k = 10
        num_queries = 20

        recall_threshold = 0.95

        # Generating sample data
        data = np.float32(np.random.random((num_elements, dim)))

        # Declaring index
        hnsw_index = hnswlib.Index(space='l2', dim=dim)  # possible options are l2, cosine or ip
        bf_index = hnswlib.BFIndex(space='l2', dim=dim)

        # Initing both hnsw and brute force indices
        # max_elements - the maximum number of elements (capacity). Will throw an exception if exceeded
        # during insertion of an element.
        # The capacity can be increased by saving/loading the index, see below.
        #
        # hnsw construction params:
        # ef_construction - controls index search speed/build speed tradeoff
        #
        # M - is tightly connected with internal dimensionality of the data. Strongly affects the memory consumption (~M)
        # Higher M leads to higher accuracy/run_time at fixed ef/efConstruction

        hnsw_index.init_index(max_elements=num_elements, ef_construction=200, M=16)
        bf_index.init_index(max_elements=num_elements)

        # Controlling the recall for hnsw by setting ef:
        # higher ef leads to better accuracy, but slower search
        hnsw_index.set_ef(200)

        # Set number of threads used during batch search/construction in hnsw
        # By default using all available cores
        hnsw_index.set_num_threads(4)

        print("Adding batch of %d elements" % (len(data)))
        hnsw_index.add_items(data)
        bf_index.add_items(data)

        print("Indices built")

        # Generating query data
        query_data = np.float32(np.random.random((num_queries, dim)))

        # Query the elements and measure recall:
        labels_hnsw, distances_hnsw = hnsw_index.knn_query(query_data, k)
        labels_bf, distances_bf = bf_index.knn_query(query_data, k)

        # Measure recall
        correct = 0
        for i in range(num_queries):
            for label in labels_hnsw[i]:
                for correct_label in labels_bf[i]:
                    if label == correct_label:
                        correct += 1
                        break

        recall_before = float(correct) / (k*num_queries)
        print("recall is :", recall_before)
        self.assertGreater(recall_before, recall_threshold)

        # test serializing  the brute force index
        index_path = 'bf_index.bin'
        print("Saving index to '%s'" % index_path)
        bf_index.save_index(index_path)
        del bf_index

        # Re-initiating, loading the index
        bf_index = hnswlib.BFIndex(space='l2', dim=dim)

        print("\nLoading index from '%s'\n" % index_path)
        bf_index.load_index(index_path)

        # Query the brute force index again to verify that we get the same results
        labels_bf, distances_bf = bf_index.knn_query(query_data, k)

        # Measure recall
        correct = 0
        for i in range(num_queries):
            for label in labels_hnsw[i]:
                for correct_label in labels_bf[i]:
                    if label == correct_label:
                        correct += 1
                        break

        recall_after = float(correct) / (k*num_queries)
        print("recall after reloading is :", recall_after)

        self.assertEqual(recall_before, recall_after)

        os.remove(index_path)



import os
import glob
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
def plot_data_from_file(file_path):
    # Load the data, assuming the last column is text
    data = pd.read_csv(file_path, header=None)
    rep_size=len(set(data[data.columns[-1]]))
    data.drop(data.columns[-1], axis=1, inplace=True)  # Drop the last column (text)

    # Number of numerical columns
    num_columns = data.shape[1]

    # Create a subplot for each column
    fig, axes = plt.subplots(num_columns, 1, figsize=(10, 6 * num_columns))
    
    # In case there is only one column, axes will not be an array, so we convert it
    if num_columns == 1:
        axes = [axes]
    
    for i, ax in enumerate(axes):
        idx=0
        ax.scatter(np.asarray(data.index,dtype=np.int64)%rep_size, data[i], label=f'Column {i+1}')
        ax.set_title(f'Column {i+1}')
        ax.set_xlabel('ID Number')
        ax.set_ylabel('Value')
        ax.legend()
        ax.grid(True)

    plt.tight_layout()
    plt.suptitle(f'Data from {os.path.basename(file_path)}')

    # Save the plot to a file
    plt.savefig(file_path.replace('.txt', '.png'))
    plt.close()

def scan_and_plot(directory):
    # Scan for .txt files in the given directory
    txt_files = glob.glob(os.path.join(directory, '*.txt'))

    # Process each file
    for file in txt_files:
        print(f'Processing {file}...')
        plot_data_from_file(file)
        print(f'Plot saved for {file}')
# Replace 'your_folder_path' with the path to the folder containing the .txt files
scan_and_plot('./')


import pickle
import unittest

import numpy as np

import hnswlib


def get_dist(metric, pt1, pt2):
    if metric == 'l2':
        return np.sum((pt1-pt2)**2)
    elif metric == 'ip':
        return 1. - np.sum(np.multiply(pt1, pt2))
    elif metric == 'cosine':
        return 1. - np.sum(np.multiply(pt1, pt2)) / (np.sum(pt1**2) * np.sum(pt2**2))**.5


def brute_force_distances(metric, items, query_items, k):
    dists = np.zeros((query_items.shape[0], items.shape[0]))
    for ii in range(items.shape[0]):
        for jj in range(query_items.shape[0]):
            dists[jj,ii] = get_dist(metric, items[ii, :], query_items[jj, :])

    labels = np.argsort(dists, axis=1) # equivalent, but faster: np.argpartition(dists, range(k), axis=1)
    dists = np.sort(dists, axis=1)     # equivalent, but faster: np.partition(dists, range(k), axis=1)

    return labels[:, :k], dists[:, :k]


def check_ann_results(self, metric, items, query_items, k, ann_l, ann_d, err_thresh=0, total_thresh=0, dists_thresh=0):
    brute_l, brute_d = brute_force_distances(metric, items, query_items, k)
    err_total = 0
    for jj in range(query_items.shape[0]):
        err = np.sum(np.isin(brute_l[jj, :], ann_l[jj, :], invert=True))
        if err > 0:
            print(f"Warning: {err} labels are missing from ann results (k={k}, err_thresh={err_thresh})")

        if err > err_thresh:
            err_total += 1

    self.assertLessEqual(err_total, total_thresh, f"Error: knn_query returned incorrect labels for {err_total} items (k={k})")

    wrong_dists = np.sum(((brute_d - ann_d)**2.) > 1e-3)
    if wrong_dists > 0:
        dists_count = brute_d.shape[0]*brute_d.shape[1]
        print(f"Warning: {wrong_dists} ann distance values are different from brute-force values (total # of values={dists_count}, dists_thresh={dists_thresh})")

    self.assertLessEqual(wrong_dists, dists_thresh, msg=f"Error: {wrong_dists} ann distance values are different from brute-force values")


def test_space_main(self, space, dim):

    # Generating sample data
    data = np.float32(np.random.random((self.num_elements, dim)))
    test_data = np.float32(np.random.random((self.num_test_elements, dim)))

    # Declaring index
    p = hnswlib.Index(space=space, dim=dim)  # possible options are l2, cosine or ip
    print(f"Running pickle tests for {p}")

    p.num_threads = self.num_threads  # by default using all available cores

    p0 = pickle.loads(pickle.dumps(p)) # pickle un-initialized Index
    p.init_index(max_elements=self.num_elements, ef_construction=self.ef_construction, M=self.M)
    p0.init_index(max_elements=self.num_elements, ef_construction=self.ef_construction, M=self.M)

    p.ef = self.ef
    p0.ef = self.ef

    p1 = pickle.loads(pickle.dumps(p)) # pickle Index before adding items

    # add items to ann index p,p0,p1
    p.add_items(data)
    p1.add_items(data)
    p0.add_items(data)

    p2=pickle.loads(pickle.dumps(p)) # pickle Index before adding items

    self.assertTrue(np.allclose(p.get_items(), p0.get_items()), "items for p and p0 must be same")
    self.assertTrue(np.allclose(p0.get_items(), p1.get_items()), "items for p0 and p1 must be same")
    self.assertTrue(np.allclose(p1.get_items(), p2.get_items()), "items for p1 and p2 must be same")

    # Test if returned distances are same
    l, d = p.knn_query(test_data, k=self.k)
    l0, d0 = p0.knn_query(test_data, k=self.k)
    l1, d1 = p1.knn_query(test_data, k=self.k)
    l2, d2 = p2.knn_query(test_data, k=self.k)

    self.assertLessEqual(np.sum(((d-d0)**2.) > 1e-3), self.dists_err_thresh, msg=f"knn distances returned by p and p0 must match")
    self.assertLessEqual(np.sum(((d0-d1)**2.) > 1e-3), self.dists_err_thresh, msg=f"knn distances returned by p0 and p1 must match")
    self.assertLessEqual(np.sum(((d1-d2)**2.) > 1e-3), self.dists_err_thresh, msg=f"knn distances returned by p1 and p2 must match")

    # check if ann results match brute-force search
    #   allow for 2 labels to be missing from ann results
    check_ann_results(self, space, data, test_data, self.k, l, d,
                           err_thresh=self.label_err_thresh,
                           total_thresh=self.item_err_thresh,
                           dists_thresh=self.dists_err_thresh)

    check_ann_results(self, space, data, test_data, self.k, l2, d2,
                           err_thresh=self.label_err_thresh,
                           total_thresh=self.item_err_thresh,
                           dists_thresh=self.dists_err_thresh)

    # Check ef parameter value
    self.assertEqual(p.ef, self.ef, "incorrect value of p.ef")
    self.assertEqual(p0.ef, self.ef, "incorrect value of p0.ef")
    self.assertEqual(p2.ef, self.ef, "incorrect value of p2.ef")
    self.assertEqual(p1.ef, self.ef, "incorrect value of p1.ef")

    # Check M parameter value
    self.assertEqual(p.M, self.M, "incorrect value of p.M")
    self.assertEqual(p0.M, self.M, "incorrect value of p0.M")
    self.assertEqual(p1.M, self.M, "incorrect value of p1.M")
    self.assertEqual(p2.M, self.M, "incorrect value of p2.M")

    # Check ef_construction parameter value
    self.assertEqual(p.ef_construction, self.ef_construction, "incorrect value of p.ef_construction")
    self.assertEqual(p0.ef_construction, self.ef_construction, "incorrect value of p0.ef_construction")
    self.assertEqual(p1.ef_construction, self.ef_construction, "incorrect value of p1.ef_construction")
    self.assertEqual(p2.ef_construction, self.ef_construction, "incorrect value of p2.ef_construction")


class PickleUnitTests(unittest.TestCase):

    def setUp(self):
        self.ef_construction = 200
        self.M = 32
        self.ef = 400

        self.num_elements = 1000
        self.num_test_elements = 100

        self.num_threads = 4
        self.k = 25

        self.label_err_thresh = 5  # max number of missing labels allowed per test item
        self.item_err_thresh = 5   # max number of items allowed with incorrect labels

        self.dists_err_thresh = 50 # for two matrices, d1 and d2, dists_err_thresh controls max
                                 # number of value pairs that are allowed to be different in d1 and d2
                                 # i.e., number of values that are (d1-d2)**2>1e-3

    def test_inner_product_space(self):
        test_space_main(self, 'ip', 16)

    def test_l2_space(self):
        test_space_main(self, 'l2', 53)

    def test_cosine_space(self):
        test_space_main(self, 'cosine', 32)



import unittest

import numpy as np

import hnswlib

class RandomSelfTestCase(unittest.TestCase):
    def testRandomSelf(self):

        data1 = np.asarray([[1, 0, 0],
                            [0, 1, 0],
                            [0, 0, 1],
                            [1, 0, 1],
                            [1, 1, 1],
                            ])

        for space, expected_distances in [
            ('l2', [[0., 1., 2., 2., 2.]]),
            ('ip', [[-2., -1., 0., 0., 0.]]),
            ('cosine', [[0, 1.835e-01, 4.23e-01, 4.23e-01, 4.23e-01]])]:

            for rightdim in range(1, 128, 3):
                for leftdim in range(1, 32, 5):
                    data2 = np.concatenate(
                        [np.zeros([data1.shape[0], leftdim]), data1, np.zeros([data1.shape[0], rightdim])], axis=1)
                    dim = data2.shape[1]
                    p = hnswlib.Index(space=space, dim=dim)
                    p.init_index(max_elements=5, ef_construction=100, M=16)

                    p.set_ef(10)

                    p.add_items(data2)

                    # Query the elements for themselves and measure recall:
                    labels, distances = p.knn_query(np.asarray(data2[-1:]), k=5)

                    
                    diff=np.mean(np.abs(distances-expected_distances))                    
                    self.assertAlmostEqual(diff, 0, delta=1e-3)



import os
import unittest

import numpy as np

import hnswlib


class RandomSelfTestCase(unittest.TestCase):
    def testRandomSelf(self):

        dim = 16
        num_elements = 10000

        # Generating sample data
        data = np.float32(np.random.random((num_elements, dim)))

        # Declaring index
        hnsw_index = hnswlib.Index(space='l2', dim=dim)  # possible options are l2, cosine or ip
        bf_index = hnswlib.BFIndex(space='l2', dim=dim)

        # Initiating index
        # max_elements - the maximum number of elements, should be known beforehand
        #     (probably will be made optional in the future)
        #
        # ef_construction - controls index search speed/build speed tradeoff
        # M - is tightly connected with internal dimensionality of the data
        #     strongly affects the memory consumption

        hnsw_index.init_index(max_elements=num_elements, ef_construction=100, M=16)
        bf_index.init_index(max_elements=num_elements)

        # Controlling the recall by setting ef:
        # higher ef leads to better accuracy, but slower search
        hnsw_index.set_ef(10)

        hnsw_index.set_num_threads(4)  # by default using all available cores

        print("Adding %d elements" % (len(data)))
        hnsw_index.add_items(data)
        bf_index.add_items(data)

        # Query the elements for themselves and measure recall:
        labels, distances = hnsw_index.knn_query(data, k=1)
        self.assertAlmostEqual(np.mean(labels.reshape(-1) == np.arange(len(data))), 1.0, 3)

        print("Querying only even elements")
        # Query the even elements for themselves and measure recall:
        filter_function = lambda id: id%2 == 0
        # Warning: search with a filter works slow in python in multithreaded mode, therefore we set num_threads=1
        labels, distances = hnsw_index.knn_query(data, k=1, num_threads=1, filter=filter_function)
        self.assertAlmostEqual(np.mean(labels.reshape(-1) == np.arange(len(data))), .5, 3)
        # Verify that there are only even elements:
        self.assertTrue(np.max(np.mod(labels, 2)) == 0)

        labels, distances = bf_index.knn_query(data, k=1, filter=filter_function)
        self.assertEqual(np.mean(labels.reshape(-1) == np.arange(len(data))), .5)



import numpy as np
import os

def normalized(a, axis=-1, order=2):
    l2 = np.atleast_1d(np.linalg.norm(a, order, axis))
    l2[l2==0] = 1
    return a / np.expand_dims(l2, axis)

N=100000
dummy_data_multiplier=3
N_queries = 1000
d=8
K=5

np.random.seed(1)

print("Generating data...")
batches_dummy= [ normalized(np.float32(np.random.random( (N,d)))) for _ in range(dummy_data_multiplier)]
batch_final = normalized (np.float32(np.random.random( (N,d))))
queries = normalized(np.float32(np.random.random( (N_queries,d))))
print("Computing distances...")
dist=np.dot(queries,batch_final.T)
topk=np.argsort(-dist)[:,:K]
print("Saving...")

try:
    os.mkdir("data")
except OSError as e:
    pass

for idx, batch_dummy in enumerate(batches_dummy):
    batch_dummy.tofile('data/batch_dummy_%02d.bin' % idx)
batch_final.tofile('data/batch_final.bin')
queries.tofile('data/queries.bin')
np.int32(topk).tofile('data/gt.bin')
with open("data/config.txt", "w") as file:
    file.write("%d %d %d %d %d" %(N, dummy_data_multiplier, N_queries, d, K))


import hnswlib
import numpy as np
import pickle


"""
Example of search
"""

dim = 128
num_elements = 10000

# Generating sample data
data = np.float32(np.random.random((num_elements, dim)))
ids = np.arange(num_elements)

# Declaring index
p = hnswlib.Index(space='l2', dim=dim)  # possible options are l2, cosine or ip

# Initializing index - the maximum number of elements should be known beforehand
p.init_index(max_elements=num_elements, ef_construction=200, M=16)

# Element insertion (can be called several times):
p.add_items(data, ids)

# Controlling the recall by setting ef:
p.set_ef(50)  # ef should always be > k

# Query dataset, k - number of the closest elements (returns 2 numpy arrays)
labels, distances = p.knn_query(data, k=1)

# Index objects support pickling
# WARNING: serialization via pickle.dumps(p) or p.__getstate__() is NOT thread-safe with p.add_items method!
# Note: ef parameter is included in serialization; random number generator is initialized with random_seed on Index load
p_copy = pickle.loads(pickle.dumps(p))  # creates a copy of index p using pickle round-trip

### Index parameters are exposed as class properties:
print(f"Parameters passed to constructor:  space={p_copy.space}, dim={p_copy.dim}")
print(f"Index construction: M={p_copy.M}, ef_construction={p_copy.ef_construction}")
print(f"Index size is {p_copy.element_count} and index capacity is {p_copy.max_elements}")
print(f"Search speed/quality trade-off parameter: ef={p_copy.ef}")



import unittest

import numpy as np

import hnswlib


class RandomSelfTestCase(unittest.TestCase):
    def testGettingItems(self):
        print("\n**** Getting the data by label test ****\n")

        dim = 16
        num_elements = 10000

        # Generating sample data
        data = np.float32(np.random.random((num_elements, dim)))
        labels = np.arange(0, num_elements)

        # Declaring index
        p = hnswlib.Index(space='l2', dim=dim)  # possible options are l2, cosine or ip

        # Initiating index
        # max_elements - the maximum number of elements, should be known beforehand
        #     (probably will be made optional in the future)
        #
        # ef_construction - controls index search speed/build speed tradeoff
        # M - is tightly connected with internal dimensionality of the data
        #     strongly affects the memory consumption

        p.init_index(max_elements=num_elements, ef_construction=100, M=16)

        # Controlling the recall by setting ef:
        # higher ef leads to better accuracy, but slower search
        p.set_ef(100)

        p.set_num_threads(4)  # by default using all available cores

        # Before adding anything, getting any labels should fail
        self.assertRaises(Exception, lambda: p.get_items(labels))

        print("Adding all elements (%d)" % (len(data)))
        p.add_items(data, labels)

        # Getting data by label should raise an exception if a scalar is passed:
        self.assertRaises(ValueError, lambda: p.get_items(labels[0]))

        # After adding them, all labels should be retrievable
        returned_items_np = p.get_items(labels)
        self.assertTrue((data == returned_items_np).all())

        # check returned type of get_items
        self.assertTrue(isinstance(returned_items_np, np.ndarray))
        returned_items_list = p.get_items(labels, return_type="list")
        self.assertTrue(isinstance(returned_items_list, list))
        self.assertTrue(isinstance(returned_items_list[0], list))



import hnswlib
import numpy as np


"""
Example of replacing deleted elements with new ones
"""

dim = 16
num_elements = 1_000
max_num_elements = 2 * num_elements

# Generating sample data
labels1 = np.arange(0, num_elements)
data1 = np.float32(np.random.random((num_elements, dim)))  # batch 1
labels2 = np.arange(num_elements, 2 * num_elements)
data2 = np.float32(np.random.random((num_elements, dim)))  # batch 2
labels3 = np.arange(2 * num_elements, 3 * num_elements)
data3 = np.float32(np.random.random((num_elements, dim)))  # batch 3

# Declaring index
hnsw_index = hnswlib.Index(space='l2', dim=dim)

# Initiating index
# max_elements - the maximum number of elements, should be known beforehand
#     (probably will be made optional in the future)
#
# ef_construction - controls index search speed/build speed tradeoff
# M - is tightly connected with internal dimensionality of the data
#     strongly affects the memory consumption

# Enable replacing of deleted elements
hnsw_index.init_index(max_elements=max_num_elements, ef_construction=200, M=16, allow_replace_deleted=True)

# Controlling the recall by setting ef:
# higher ef leads to better accuracy, but slower search
hnsw_index.set_ef(10)

# Set number of threads used during batch search/construction
# By default using all available cores
hnsw_index.set_num_threads(4)

# Add batch 1 and 2 data
hnsw_index.add_items(data1, labels1)
hnsw_index.add_items(data2, labels2)  # Note: maximum number of elements is reached

# Delete data of batch 2
for label in labels2:
    hnsw_index.mark_deleted(label)

# Replace deleted elements
# Maximum number of elements is reached therefore we cannot add new items,
# but we can replace the deleted ones by using replace_deleted=True
hnsw_index.add_items(data3, labels3, replace_deleted=True)
# hnsw_index contains the data of batch 1 and batch 3 only



import os.path
import os

links = ['ftp://ftp.irisa.fr/local/texmex/corpus/bigann_query.bvecs.gz',
         'ftp://ftp.irisa.fr/local/texmex/corpus/bigann_gnd.tar.gz',
         'ftp://ftp.irisa.fr/local/texmex/corpus/bigann_base.bvecs.gz']

os.makedirs('downloads', exist_ok=True)
os.makedirs('bigann', exist_ok=True)
for link in links:
    name = link.rsplit('/', 1)[-1]
    filename = os.path.join('downloads', name)
    if not os.path.isfile(filename):
        print('Downloading: ' + filename)
        try:
            os.system('wget --output-document=' + filename + ' ' + link)
        except Exception as inst:
            print(inst)
            print('  Encountered unknown error. Continuing.')
    else:
        print('Already downloaded: ' + filename)
    if filename.endswith('.tar.gz'):
        command = 'tar -zxf ' + filename + ' --directory bigann'
    else:
        command = 'cat ' + filename + ' | gzip -dc > bigann/' + name.replace(".gz", "")
    print("Unpacking file:", command)
    os.system(command)



import unittest

import numpy as np

import hnswlib


class RandomSelfTestCase(unittest.TestCase):
    def testBFIndex(self):

        dim = 16
        num_elements = 10000
        num_queries = 1000
        k = 20

        # Generating sample data
        data = np.float32(np.random.random((num_elements, dim)))

        # Declaring index
        bf_index = hnswlib.BFIndex(space='l2', dim=dim)  # possible options are l2, cosine or ip
        bf_index.init_index(max_elements=num_elements)

        num_threads = 8
        bf_index.set_num_threads(num_threads)  # by default using all available cores

        print(f"Adding all elements {num_elements}")
        bf_index.add_items(data)

        self.assertEqual(bf_index.num_threads, num_threads)
        self.assertEqual(bf_index.get_max_elements(), num_elements)
        self.assertEqual(bf_index.get_current_count(), num_elements)

        queries = np.float32(np.random.random((num_queries, dim)))
        print("Searching nearest neighbours")
        labels, distances = bf_index.knn_query(queries, k=k)

        print("Checking results")
        for i in range(num_queries):
            query = queries[i]
            sq_dists = (data - query)**2
            dists = np.sum(sq_dists, axis=1)
            labels_gt = np.argsort(dists)[:k]
            dists_gt = dists[labels_gt]
            dists_bf = distances[i]
            # we can compare labels but because of numeric errors in distance calculation in C++ and numpy
            # sometimes we get different order of labels, therefore we compare distances
            max_diff_with_gt = np.max(np.abs(dists_gt - dists_bf))

            self.assertTrue(max_diff_with_gt < 1e-5)



import hnswlib
import numpy as np


"""
Example of filtering elements when searching
"""

dim = 16
num_elements = 10000

# Generating sample data
data = np.float32(np.random.random((num_elements, dim)))

# Declaring index
hnsw_index = hnswlib.Index(space='l2', dim=dim)  # possible options are l2, cosine or ip

# Initiating index
# max_elements - the maximum number of elements, should be known beforehand
#     (probably will be made optional in the future)
#
# ef_construction - controls index search speed/build speed tradeoff
# M - is tightly connected with internal dimensionality of the data
#     strongly affects the memory consumption

hnsw_index.init_index(max_elements=num_elements, ef_construction=100, M=16)

# Controlling the recall by setting ef:
# higher ef leads to better accuracy, but slower search
hnsw_index.set_ef(10)

# Set number of threads used during batch search/construction
# By default using all available cores
hnsw_index.set_num_threads(4)

print("Adding %d elements" % (len(data)))
# Added elements will have consecutive ids
hnsw_index.add_items(data, ids=np.arange(num_elements))

print("Querying only even elements")
# Define filter function that allows only even ids
filter_function = lambda idx: idx%2 == 0
# Query the elements for themselves and search only for even elements:
# Warning: search with a filter works slow in python in multithreaded mode, therefore we set num_threads=1
labels, distances = hnsw_index.knn_query(data, k=1, num_threads=1, filter=filter_function)
# labels contain only elements with even id



import os
import sys
import platform

import numpy as np
import pybind11
import setuptools
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

__version__ = '0.8.0'


include_dirs = [
    pybind11.get_include(),
    np.get_include(),
]

# compatibility when run in python_bindings
bindings_dir = 'python_bindings'
if bindings_dir in os.path.basename(os.getcwd()):
    source_files = ['./bindings.cpp']
    include_dirs.extend(['../hnswlib/'])
else:
    source_files = ['./python_bindings/bindings.cpp']
    include_dirs.extend(['./hnswlib/'])


libraries = []
extra_objects = []


ext_modules = [
    Extension(
        'hnswlib',
        source_files,
        include_dirs=include_dirs,
        libraries=libraries,
        language='c++',
        extra_objects=extra_objects,
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.
    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    compiler_flag_native = '-march=native'
    c_opts = {
        'msvc': ['/EHsc', '/openmp', '/O2'],
        'unix': ['-O3', compiler_flag_native],  # , '-w'
    }
    link_opts = {
        'unix': [],
        'msvc': [],
    }

    if os.environ.get("HNSWLIB_NO_NATIVE"):
        c_opts['unix'].remove(compiler_flag_native)

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']
        link_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']
    else:
        c_opts['unix'].append("-fopenmp")
        link_opts['unix'].extend(['-fopenmp', '-pthread'])

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = BuildExt.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
            if not os.environ.get("HNSWLIB_NO_NATIVE"):
                # check that native flag is available
                print('checking avalability of flag:', BuildExt.compiler_flag_native)
                if not has_flag(self.compiler, BuildExt.compiler_flag_native):
                    print('removing unsupported compiler flag:', BuildExt.compiler_flag_native)
                    opts.remove(BuildExt.compiler_flag_native)
                    # for macos add apple-m1 flag if it's available
                    if sys.platform == 'darwin':
                        m1_flag = '-mcpu=apple-m1'
                        print('checking avalability of flag:', m1_flag)
                        if has_flag(self.compiler, m1_flag):
                            print('adding flag:', m1_flag)
                            opts.append(m1_flag)
                        else:
                            print(f'flag: {m1_flag} is not available')
                else:
                    print(f'flag: {BuildExt.compiler_flag_native} is available')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())

        for ext in self.extensions:
            ext.extra_compile_args.extend(opts)
            ext.extra_link_args.extend(BuildExt.link_opts.get(ct, []))

        build_ext.build_extensions(self)


setup(
    name='hnswlib',
    version=__version__,
    description='hnswlib',
    author='Yury Malkov and others',
    url='https://github.com/yurymalkov/hnsw',
    long_description="""hnsw""",
    ext_modules=ext_modules,
    install_requires=['numpy'],
    cmdclass={'build_ext': BuildExt},
    zip_safe=False,
)



import hnswlib
import numpy as np
import threading
import pickle


"""
Example of python wrapper for hnswlib that supports python objects as ids
"""

class Index():
    def __init__(self, space, dim):
        self.index = hnswlib.Index(space, dim)
        self.lock = threading.Lock()
        self.dict_labels = {}
        self.cur_ind = 0

    def init_index(self, max_elements, ef_construction=200, M=16):
        self.index.init_index(max_elements=max_elements, ef_construction=ef_construction, M=M)

    def add_items(self, data, ids=None):
        if ids is not None:
            assert len(data) == len(ids)
        num_added = len(data)
        with self.lock:
            start = self.cur_ind
            self.cur_ind += num_added
        int_labels = []

        if ids is not None:
            for dl in ids:
                int_labels.append(start)
                self.dict_labels[start] = dl
                start += 1
        else:
            for _ in range(len(data)):
                int_labels.append(start)
                self.dict_labels[start] = start
                start += 1
        self.index.add_items(data=data, ids=np.asarray(int_labels))

    def set_ef(self, ef):
        self.index.set_ef(ef)

    def load_index(self, path):
        self.index.load_index(path)
        with open(path + ".pkl", "rb") as f:
            self.cur_ind, self.dict_labels = pickle.load(f)

    def save_index(self, path):
        self.index.save_index(path)
        with open(path + ".pkl", "wb") as f:
            pickle.dump((self.cur_ind, self.dict_labels), f)

    def set_num_threads(self, num_threads):
        self.index.set_num_threads(num_threads)

    def knn_query(self, data, k=1):
        labels_int, distances = self.index.knn_query(data=data, k=k)
        labels = []
        for li in labels_int:
            labels.append(
                [self.dict_labels[l] for l in li]
            )
        return labels, distances
